import csv
from sympy import symbols, lambdify
import numpy as np
import math
import copy

def approxinterp_read_csv(filename):
    """
    Функция считывания данных из файла .csv 

    Parameters
    ----------
    filename : str
        Абсолютный или локальный путь к файлу.

    Returns
    -------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    n : int
        Длина выборки координат.

    """
    
    with open(filename, 'r') as file:
        data = csv.reader(file, delimiter=',')
        data = list(data)
        y_list = []
        x_list = []
        for i in range(len(data)):
            x_list.append(eval(data[i][0]))
            y_list.append(eval(data[i][1]))
    x_list = np.array(x_list)
    y_list = np.array(y_list)
    n = len(x_list)
    return x_list, y_list, n


def polinom_chebyshev(x_list, y_list, n):
    """
    Фильтрация выборки координа методом полинома Чебышева

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    n : int
        Длина выборки координат.

    Returns
    -------
    new_x_list : [float, float, ...]
        Новый список координат х.
    new_y_list : [float, float, ...]
        Новый список координат у.
    new_n : int
        Новая длина выборки координат.

    """
    
    new_n = n // 8
    new_index = sorted(list(set([round((np.math.cos((2 * k + 1) * np.math.pi / (2 * new_n + 2)) + 1) * (n - 1) / 2) for k in range(new_n)][::-1])))
    new_x_list = x_list[new_index]
    new_y_list = y_list[new_index]
    return new_x_list, new_y_list, new_n


def lagranje(x_list, y_list):
    """
    Интерполяция выборки методом Лагранжа

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.

    Returns
    -------
    new_L : function
        Интерполированная функция для введённой выборки.

    """
    
    x = symbols('x')
    L = 0
    l = []
    for j in range(len(x_list)):
        l.append(1)
        for k in range(len(x_list)):
            if k!=j:
                l[j] *= ((x-x_list[k])/(x_list[j]-x_list[k]))
        L += y_list[j]*l[j]
    new_L = lambdify(x, L)
    return new_L



def newton(x_list, y_list, interp_type=True):
    """
    Интерполяция выборки методом Ньютона

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    interp_type : bool, optional
        Тип интерполяции (вперёд\назад). 
        По умолчанию True - вперёд.
        False - назад.

    Raises
    ------
    ValueError
        Возникает при передаче не равноотстоящих узлов.

    Returns
    -------
    func : function
        Интерполированная функция для введённой выборки.

    """
    
    for i in range(len(x_list)-2):
        if abs((x_list[i+1] - x_list[i]) - (x_list[i+2] - x_list[i+1])) > 0.1 ** 12:
            raise ValueError('Данный метод применим только для равноотстоящих узлов!')
             
    #конечные разности
    d = []
    d.append(y_list)
    for i in range(len(y_list)-1):
        temp = d[i]
        d.append([])
        for j in range(len(temp)-1):
            d[i+1].append(temp[j+1]-temp[j])
        
    if interp_type: 
        #интерполяция вперед 
        x = symbols('x')
        h = x_list[1] - x_list[0]
        n = len(x_list)-1
        qv = (x - x_list[0])/h
        Pv = 0
        qspisv = []
        qspisv.append(1)
        for i in range(n):
            tmep = qspisv[i]
            qspisv.append(tmep*(qv-i))
    
        for i in range(n+1):
            Pv += qspisv[i]*d[i][0]/np.math.factorial(i)
    
        func = lambdify(x, Pv)

    else:
        #интерполяция назад
        x = symbols('x')
        h = x_list[1] - x_list[0]
        n = len(x_list)-1
        qn = (x - x_list[-1])/h
        Pn = 0
        qspisn = []
        qspisn.append(1)
        for i in range(n):
            tmep = qspisn[i]
            qspisn.append(tmep*(qn+i))
    
        for i in range(n+1):
            Pn += qspisn[i]*d[i][-1]/np.math.factorial(i)
    
        func = lambdify(x, Pn)
    return func



def jdmethod_fast(a):
    """
    Решение СЛАУ методом Жордана-Гаусса, используя numpy

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введённая пользователем.

    Returns
    -------
    [float, float, ...]
        Решение СЛАУ.

    """
    
    a = np.array(a, float)
    n = len(a)
    for k in range(n):
        if np.fabs(a[k,k]) < 1.0e-12:
            for i in range(k+1,n):
                if np.fabs(a[i,k]) > np.fabs(a[k,k]):
                    a[[i, k]] = a[[k, i]]
                    break
        pivot = a[k,k]
        a[k] = a[k] / pivot

        a[np.arange(n)!=k] -= a[np.arange(n)!=k][:, k].reshape((n-1, 1)) * a[k]
    return a[:, -1:]


def qubic_splines(x_list, y_list):
    """
    

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.

    Returns
    -------
    S : [function, function, ...]
        Список функций для каждого сплайна на указанном отрезке.

    """
    x = symbols('x')
    S = []
    a = y_list[:]
    b = []
    d = []
    number_of_splines = len(x_list)-1
    h = np.array(x_list[1:]) - np.array(x_list[:-1])
    C_to_solve = np.zeros((number_of_splines + 1, number_of_splines + 2))
    for j in range(number_of_splines + 1):
        if j == 0 or j == number_of_splines:
            C_to_solve[j][j] = 1
        else:
            C_to_solve[j][j - 1] = h[j - 1] / 3
            C_to_solve[j][j] = 2 * (h[j - 1] + h[j]) / 3
            C_to_solve[j][j + 1] = h[j] / 3
            C_to_solve[j][-1] = (y_list[j + 1] - y_list[j]) / h[j] - (y_list[j] - y_list[j - 1]) / h[j - 1]
    c = jdmethod_fast(C_to_solve).transpose()[0]
    for i in range(number_of_splines):
        d.append((c[i+1] - c[i])/(3*h[i]))
        b.append(((y_list[i+1]-y_list[i])/h[i]) - h[i]*(c[i+1]+2*c[i])/3)
        S.append(a[i] + b[i]*(x-x_list[i]) + c[i]*(x-x_list[i])*(x-x_list[i]) + d[i]*(x-x_list[i])*(x-x_list[i])*(x-x_list[i]))
    S.append(S[-1])
    return S


def linear_approximation(x_list, y_list):
    """
    Линейная аппроксимация выборки

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.

    Returns
    -------
    k : float
        Коэффициент k в уравнении y=kx+b 
    b : float
        Коэффициент b в уравнении y=kx+b 

    """
    summ1 = 0
    n = len(x_list)
    for i in range(len(x_list)):
        summ1 += x_list[i]*y_list[i]     
    summ2 = sum(x_list)*sum(y_list)
    k = (n*summ1 - summ2)/(n*sum([i**2 for i in x_list])-sum(x_list)**2)
    b = (sum(y_list)-k*sum(x_list))/n
    return k,b 


def dispersion_linear(x_list, y_list, k, b):
    """
    Функция находит дисперсию для линейной аппроксимации

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    k : float
        Коэффициент k в уравнении y=kx+b 
    b : float
        Коэффициент b в уравнении y=kx+b 

    Returns
    -------
    summ_d : float
        Дисперсия линейной аппроксимации

    """
    
    x_values = np.linspace(x_list[0], x_list[-1], num=len(x_list))
    y_values = np.poly1d([k, b])(x_values)
    summ_d = 0
    for i in range(len(x_list)):
        summ_d = (y_values[i] - y_list[i])**2
    return summ_d


def quadratic_approximation(x_list, y_list):
    """
    Аппроксимация квадратичной функцией

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.

    Returns
    -------
    x_koef : [float, float, float]
        Список из коэффициентов k1,k2,k3 в уравнении y=k1*x^2+k2*x+k3

    """
    n = len(x_list)
    summ_kv = 0
    for i in range(len(x_list)):
        summ_kv+= x_list[i]**2
    summ_cube = 0
    for i in range(len(x_list)):
        summ_cube+= x_list[i]**3
    summ_four = 0
    for i in range(len(x_list)):
        summ_four+= x_list[i]**4
    summ_xy = 0
    for i in range(len(x_list)):
        summ_xy += x_list[i]*y_list[i]
    summ_xkvy = 0
    for i in range(len(x_list)):
        summ_xkvy += (x_list[i]**2)*y_list[i]
    
    A = [[summ_kv, sum(x_list), n, sum(y_list)], [summ_cube, summ_kv, sum(x_list), summ_xy], [summ_four, summ_cube, summ_kv, summ_xkvy]]
    x_koef = jdmethod_fast(A).transpose()[0]
    return x_koef



def dispersion_quadratic(x_list, y_list, x_koef):
    """
    Функця находит дисперсию для квадратичной аппроксимации

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    x_koef : [float, float, float]
        Список из коэффициентов k1,k2,k3 в уравнении y=k1*x^2+k2*x+k3

    Returns
    -------
    summ_d : float
        Дисперсия линейной аппроксимации

    """
    x_values = np.linspace(x_list[0], x_list[-1], num=len(x_list))
    y_values = np.poly1d(x_koef)(x_values)
    summ_d = 0
    for i in range(len(x_list)):
        summ_d = (y_values[i] - y_list[i])**2
    return summ_d



def Gauss_approximation(x_list):
    """
    Аппроксимация методом Гаусса для формулы
    y_i = a * exp(-(x_i-b)^2/c^2)

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.

    Returns
    -------
    a : float
        Коэф а.
    b : float
        Коэф b.
    c : float
        Коэф c.
    """
    n = len(x_list)
    x__ = 1/n*(sum(x_list))
    p = 1/200
    summ_ = 0
    for i in range(len(x_list)):
        summ_ += (x_list[i] - x__)**2
    S_0 = math.sqrt((1/(n-1))*summ_)
    c = S_0/(math.sqrt(S_0))
    summ_E = 0
    for i in range(len(x_list)): 
        summ_E += x_list[i]*p
    b = summ_E
    a = 1/(c*(math.sqrt(2*math.pi)))
    return a, b, c



def dispersion_Gauss(x_list, y_list, a, b, c):
    """
    Дисперсия для аппроксимации методом Гаусса

    Parameters
    ----------
    x_list : [float, float, ...]
        Список координат х.
    y_list : [float, float, ...]
        Список координат у.
    a : float
        Коэф а уравненяи.
    b : float
        Коэф b уравненяи.
    c : float
        Коэф c уравненяи.

    Returns
    -------
    summ_d : float
        Дисперсия линейной аппроксимации

    """
    x_values = np.linspace(x_list[0], x_list[-1], num=len(x_list))
    y_values = [a*math.e**(-((x_i-b)**2)/(c**2)) for x_i in x_values]
    summ_d = 0
    for i in range(len(x_list)):
        summ_d = (y_values[i] - y_list[i])**2
    return summ_d


from sympy import *
import numpy as np
import matplotlib.pyplot as plt

def draw_plot(values, label, three_d=False):
    """
    Функция отрисовки графика функции по её передаваемым значениям

    Parameters
    ----------
    values : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.
    label : str
        Подпись функции на графике.
    three_d : bool, optional
        Отрисовка графика в 3D. 
        По умолчанию False.

    Raises
    ------
    ValueError
        Возникает при попытке нарисовать 3D график используя только 2 переменные.
        
    """
    if not three_d or len(values) < 3:
        if three_d:
            raise ValueError('Предупреждение, вы не можете построить 3D график используя только две переменные')
        try:
            x_list = values[:, 1]
            y_list = values[:, 2]
        except:
            x_list = values[0]
            y_list = values[1]
        plt.plot(x_list, y_list, label=label)
        plt.legend()
        plt.show()
    else:
        try:
            x_list = values[:, 1]
            y_list = values[:, 2]
            z_list = values[:, 3]
        except:
            x_list = values[0]
            y_list = values[1]
            z_list = values[2]
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        ax.plot(x_list, y_list, z_list, label=label)
        ax.legend(loc='upper left')
        plt.show()
        

def draw_plots(values, labels, colors=[], three_d=False):
    """
    

    Parameters
    ----------
    values : [[[float, float, ...],
              [float, float, ...],
              ...],
              [[float, float, ...],
              [float, float, ...],
              ...], ...]
        Список из нескольких списков из групп значений x, y, y', y'' и тд. (Несколько графиков)
    labels : [str, str, ...]
        Список подписей функций на графике.
    colors : [str, str, ...], optional
        Список цветов отрисовки графиков. 
        По умолчанию [].
    three_d : bool, optional
        Отрисовка графика в 3D. 
        По умолчанию False.

    """
    
    if not three_d:
        for i in range(len(values)):
            try:
                x_list = values[i][:, 1]
                y_list = values[i][:, 2]
            except:
                x_list = values[i][0]
                y_list = values[i][1]
            if colors:
                plt.plot(x_list, y_list, colors[i], label=labels[i])
            else:
                plt.plot(x_list, y_list, label=labels[i])
        plt.legend()
        plt.show()
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        for i in range(len(values)):   
            try:
                x_list = values[i][:, 1]
                y_list = values[i][:, 2]
                z_list = values[i][:, 3]
            except:
                x_list = values[i][0]
                y_list = values[i][1]
                z_list = values[i][2]
            if colors:
                ax.plot(x_list, y_list, z_list, color=colors[i], label=labels[i])
            else:
                ax.plot(x_list, y_list, z_list, label=labels[i])                                                 
        ax.legend(loc='upper left')
        plt.show()



def input_expression():
    """
    Возвращает введённую функцию\функции 
    
    Returns
    -------
    expressions : optional - (exp1, exp2) or (exp, )
        Кортеж из функции\системы двух функции.
    n : int
        Колв-во  введённых функций.

    """
    
    n = int(input('Введите кол-во уравнений, которые хотите задать (1|2)\n ->'))
    if n == 1:
        expression = input('Введите уравнение\n ->')
        return (expression, ), 1
    exp1 = input("Введите первое уравнение\n ->y'=")
    exp2 = input("Введите второе уравнение\n ->z'=")
    return (exp1, exp2), 2  

def lambda_func(string, variables):
    """
    Генерирует на основании введённого уравнения - его представление в функциях

    Parameters
    ----------
    string : srt
        Введённое уравнение.
    variables : [str, str, ...]
        Список переменных, встречающихся в уравнении.

    Returns
    -------
    func: function
        Лямбда-функция, основанная на введённом уравнении.

    """
    var = []
    for v in variables:
        exec(f'{v} = Symbol("{v}")')
        eval(f'var.append({v})')
    expr = eval(string.replace('^', '**'))
    return lambdify(var, expr)
    
def parse_string(string):
    """
    Парсер, приводящий любое уравнение к виду y'(n) = x + y' + y'' + y''' + ... + y'(n-1)

    Parameters
    ----------
    string : str
        Введённое уравнение.

    Returns
    -------
    exp : str
        Преобразованное уравнение.
    max_val : int
        Степень диф. уравнения.
    variables : [str, str, ...]
        Список участвующих в уравнении переменных.

    """
    
    string = string.replace(' ', '')
    expression = []
    variable = ''
    for i in range(len(string)):
        variable += string[i]
        if i == len(string) - 1:
            expression.append(variable)
            variable = ''
        elif not (string[i + 1] in "'^" or string[i + 1].isalpha() or string[i + 1].isdigit()) or string[i] in '=+*/-':
            expression.append(variable)
            variable = ''
    max_ind = 0
    max_val = -1
    for var in expression:
        if var.count("'") > max_val:
            max_val = var.count("'")
            max_ind = expression.index(var)
    val = expression[max_ind]
    eq = expression.index('=')
    if eq < max_ind:
        expression = expression[eq + 1:] + ['='] + expression[:eq]
    eq = expression.index('=')
    left = expression[:eq]
    if left[0] not in '+-' and ((val in left) == (len(left) > 2)) or len(left) == 1:
        left = ['+'] + left
    right = expression[eq + 1:]
    if right[0] not in '+-' and ((val in right) == (len(right) > 2)) or len(left) == 1:
        right = ['+'] + right
    new_ind = left.index(val)
    sign = left[new_ind - 1]
    left.pop(new_ind)
    left.pop(new_ind - 1)
    if sign == '+' and len(left) != 0:
        expression = right + ['-', '(']  + left + [')']
    elif sign == '+':
        expression = right
    elif sign == '-' and len(right) != 0:
        expression = left + ['-', '(']  + right + [')']
    else:
        expression = left
    if expression[0] == '+':
        expression.pop(0)
    variables = []
    for v in expression:
        temp = []
        for s in v:
            temp.append(s.isalpha())
        if sum(temp):
            new_temp = []
            for s in v:
                if s.isalpha() or s == "'":
                    new_temp.append(s)
            variables.append(''.join(new_temp))
    variables = sorted(variables)
    variables = sorted(variables, key=lambda x: len(x))
    print(f"Функция : {' '.join(expression)}\nСтепень дифф. уравнения: {max_val}")
    return ' '.join(expression), max_val, variables


def eyler_cauchy(f, power, var, n=10000):
    """
    Решение диф. уравнение методом Эйлера-Коши

    Parameters
    ----------
    f : str
        Введённая функция.
    power : int
        Степень диф. уравнения.
    var : [str, str, ...]
        Список переменных.
    n : int, optional
        Кол-во получаемых значений. 
        По умолчанию 10000.

    Returns
    -------
    result : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.

    """
    
    for i in range(len(var)):
        if "'" in var[i]:
            var[i] = "y" + str(var[i].count("'") - 1)
    for v in var[::-1]:
        if v[-1].isdigit():
            f = f.replace('y' + "'" * (int(v[-1]) + 1), v)
            
    func = lambda_func(f, var)
    values_dict = {}
    for i in range(len(var)):
        values_dict[(i, var[i])] = [float(input(f'Введите начальные условия ({var[i]}_0)'))]
    
    a = float(input('Введите начальные условия (a)'))
    b = float(input('Введите начальные условия (b)'))
    h = (b-a)/n

    for i in range(n):
        values_i = []
        values_i1 = []
        for v in values_dict.keys():
            if v[0] == 0:
                values_dict[v].append(values_dict[v][i] + h)
                values_i.append(values_dict[v][i])
                values_i1.append(values_dict[v][i + 1])
            elif v[0] != len(values_dict.keys()) - 1:
                values_dict[v].append(values_dict[v][i] + h * values_dict[(v[0] + 1, var[v[0] + 1])][i])
                values_i.append(values_dict[v][i])
                values_i1.append(values_dict[v][i + 1])
            else:
                values_i.append(values_dict[v][i])
                values_i1.append(values_dict[v][i] + h * func(*values_i))
                values_dict[v].append(values_dict[v][i] + h / 2 * (func(*values_i) + func(*values_i1)))
    result = []
    for i in range(n + 1):
        result.append([])
        result[i].append(i)
        for v in values_dict.keys():
            result[i].append(values_dict[v][i])
        result[i].append(func(*result[i][1:]))
    return np.array(result)

def runge_kutti(f, power, var, n=10000):
    """
    Решение диф. уравнение методом Рунге-Кутти

    Parameters
    ----------
    f : str
        Введённая функция.
    power : int
        Степень диф. уравнения.
    var : [str, str, ...]
        Список переменных.
    n : int, optional
        Кол-во получаемых значений. 
        По умолчанию 10000.

    Returns
    -------
    result : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.

    """
    for i in range(len(var)):
        if "'" in var[i]:
            var[i] = "y" + str(var[i].count("'") - 1)
    for v in var[::-1]:
        if v[-1].isdigit():
            f = f.replace('y' + "'" * (int(v[-1]) + 1), v)
    func = lambda_func(f, var)
    values_dict = {}
    for i in range(len(var)):
        values_dict[(i, var[i])] = [float(input(f'Введите начальные условия ({var[i]}_0)'))]
    
    a = float(input('Введите начальные условия (a)'))
    b = float(input('Введите начальные условия (b)'))
    h = (b-a)/n
    
    koefs = np.zeros((len(var), 4))
    koefs[0] = np.array([0, h, h, h])
    funcs = {}
    variables = []
    for v in var:
        exec(f'{v} = Symbol("{v}")')
        eval(f'variables.append({v})')
        
    for v in values_dict.keys():
        if v[0] != 0 and v[0] != len(values_dict.keys()) - 1:
            funcs[v[0]] = lambdify(variables, var[v[0] + 1])
        elif v[0] == len(values_dict.keys()) - 1:
            funcs[v[0]] = func
    for i in range(n):
        values_dict[(0, 'x')].append(values_dict[(0, 'x')][i] + h)
        values_i = []
        for v1 in values_dict.keys():
            values_i.append(values_dict[v1][i])
        values_i = np.array(values_i)
        for j in range(1, len(values_i)):
            koefs[j][0] = h * funcs[j](*values_i)
        for j in range(1, len(values_i)):
            koefs[j][1] = h * funcs[j](*(values_i + koefs[:, 0]/2))
        for j in range(1, len(values_i)):
            koefs[j][2] = h * funcs[j](*(values_i + koefs[:, 1]/2))
        for j in range(1, len(values_i)):
            koefs[j][3] = h * funcs[j](*(values_i + koefs[:, 2]))
        
        for v in values_dict.keys():
            if v[0] != 0:
                values_dict[v].append(values_dict[v][i] + 1/6*(koefs[v[0]][0] + 2 * koefs[v[0]][1] + 2 * koefs[v[0]][2] + koefs[v[0]][3]))
        
    result = []
    for i in range(n + 1):
        result.append([])
        result[i].append(i)
        for v in values_dict.keys():
            result[i].append(values_dict[v][i])
        result[i].append(func(*result[i][1:]))
    return np.array(result)

def eyler_cauchy_system(expressions, var=['x', 'y', 'z'], n=1000):
    """
    Решение системы из двух диф. уравнений методом Эйлера-Коши

    Parameters
    ----------
    expressions : [str, str]
        Список из двух введённых функций.
    var : [str, str, str], optional
        Список переменных.
        По умолчанию ['x', 'y', 'z']
    n : int, optional
        Кол-во получаемых значений. 
        По умолчанию 10000.

    Returns
    -------
    result : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.

    """
    funcs = {'y': lambda x, y, z: eval(expressions[0].replace('^', '**')),
             'z': lambda x, y, z: eval(expressions[0].replace('^', '**'))}
    x0 = float(input(f'Введите начальные условия (x0)\n ->'))
    y0 = float(input(f'Введите начальные условия (y0)\n ->'))
    z0 = float(input(f'Введите начальные условия (z0)\n ->'))
    a = float(input('Введите начальные условия (a)'))
    b = float(input('Введите начальные условия (b)'))
    h = (b-a)/n
    values_dict = {'x': [x0],
                   'y': [y0],
                   'z': [z0]}
    
    for i in range(n):
        values_dict['x'].append(values_dict['x'][i] + h)
        values_i = []
        for v1 in values_dict.keys():
            values_i.append(values_dict[v1][i])
        yw1, zw1 = values_dict['y'][i] + h * funcs['y'](*values_i),\
                   values_dict['z'][i] + h * funcs['z'](*values_i)
        values_i1 = [values_dict['x'][i + 1], yw1, zw1]
        values_dict['y'].append(values_dict['y'][i] + h / 2 * (funcs['y'](*values_i) + funcs['y'](*values_i1)))
        values_dict['z'].append(values_dict['z'][i] + h / 2 * (funcs['z'](*values_i) + funcs['z'](*values_i1)))
    
    result = []
    for i in range(n + 1):
        result.append([])
        result[i].append(i)
        for v in values_dict.keys():
            result[i].append(values_dict[v][i])
    return np.array(result)

def runge_kutti_system(expressions, var=['x', 'y', 'z'], n=1000):
    """
    Решение системы из двух диф. уравнений методом Рунге-Кутти

    Parameters
    ----------
    expressions : [str, str]
        Список из двух введённых функций.
    var : [str, str, str], optional
        Список переменных.
        По умолчанию ['x', 'y', 'z']
    n : int, optional
        Кол-во получаемых значений. 
        По умолчанию 10000.

    Returns
    -------
    result : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.

    """
    funcs = {'y': lambda x, y, z: eval(expressions[0].replace('^', '**')),
             'z': lambda x, y, z: eval(expressions[0].replace('^', '**'))}
    x0 = float(input(f'Введите начальные условия (x0)\n ->'))
    y0 = float(input(f'Введите начальные условия (y0)\n ->'))
    z0 = float(input(f'Введите начальные условия (z0)\n ->'))
    a = float(input('Введите начальные условия (a)'))
    b = float(input('Введите начальные условия (b)'))
    h = (b-a)/n
    values_dict = {'x': [x0],
                   'y': [y0],
                   'z': [z0]}
    
    for i in range(n):
        values_i = []
        for v1 in values_dict.keys():
            values_i.append(values_dict[v1][i])
        values_i = np.array(values_i)
        k1 = h * funcs['y'](*values_i)
        l1 = h * funcs['z'](*values_i)
        k2 = h * funcs['y'](*(values_i + np.array([h/2, k1/2, l1/2])))
        l2 = h * funcs['z'](*(values_i + np.array([h/2, k1/2, l1/2])))
        k3 = h * funcs['y'](*(values_i + np.array([h/2, k2/2, l2/2])))
        l3 = h * funcs['z'](*(values_i + np.array([h/2, k2/2, l2/2])))
        k4 = h * funcs['y'](*(values_i + np.array([h, k3, l3])))
        l4 = h * funcs['z'](*(values_i + np.array([h, k3, l3])))
        
        values_dict['x'].append(values_dict['x'][i] + h)
        values_dict['y'].append(values_dict['y'][i] + 1 / 6 * (k1 + 2 * k2 + 2 * k3 + k4))
        values_dict['z'].append(values_dict['z'][i] + 1 / 6 * (l1 + 2 * l2 + 2 * l3 + l4))
    
    result = []
    for i in range(n + 1):
        result.append([])
        result[i].append(i)
        for v in values_dict.keys():
            result[i].append(values_dict[v][i])
    return np.array(result)
    
def solve_expressions(expressions, function):
    """
    Решает диф. уравнение или систему диф. уравнение

    Parameters
    ----------
    expressions : optional (str, ) or (str1, str2)
        Введённые диф. уравнения.
    function : function
        Метод для решения диф. уравнений.

    Returns
    -------
    result : [[float, float, ...],
              [float, float, ...],
              ...]
        Список из списков групп значений x, y, y', y'' и тд.

    """
    if len(expressions) == 1:
        f, power, v = parse_string(expressions[0])
        res = function(f, power, v, n=100)
        return res
    return function(expressions)
    


import csv
import numpy as np
import copy
from fractions import Fraction

def slau_inp_matr():
    """
    Функция возвращает матрицу, введённую пользователем с клавиатуры.

    Returns
    -------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    """
    while True:
        try:
            m = int(input('Сколько будет строк в матрице? '))
        except:
            print('Вы ввели не число')
        else:
            if m > 0:
                break
            else:
                print('Вы ввели не натуральное число') 

    while True:
        try:
            n = int(input('Сколько будет столбцов в матрице? '))
        except:
            print('Вы ввели не число')
        else:
            if n > 0:
                break
            else:
                print('Вы ввели не натуральное число')
                
    print("Введите элементы матрицы (заполнение идёт по строкам)")
    a = []
    for i in range(m):
        a.append([])
        for j in range(n):
            while True:
                try:
                    print(f'Введите элемент a[{i+1}][{j+1}]')
                    elem = eval(input())
                except:
                    print('Вы ввели не число')
                else:
                    break
            a[i].append(elem)
    return a



def csv_inp_matr(filename):
    """
    Функция импортирует матрицу из csv

    Parameters
    ----------
    filename : str
        Абсолютный или локальный путь до файла .csv

    Returns
    -------
    data : [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, импортированная из csv

    """
    with open(filename, 'r') as f:
        reader = csv.reader(f)
        data = list(reader)
    for i in range(len(data)):
        for j in range(len(data[0])):
            data[i][j] = eval(data[i][j])
    return data
    

def slau_outp_matr(a):
    """
    Функция для вывода матрицы

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем


    """
    for i in range(len(a)): 
        print(f'{a[i]}')      

def det(a):
    """
    Функция вычисляет определитель матрицы.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, определитель которой нужно вычислить.

    Raises
    ------
    ValueError
        Возникает, если введённая матрица a не квадратная

    Returns
    -------
    summ : float
        Определитель матрицы

    """
    if len(a) != len(a[0]):
        raise ValueError('Матрица должна быть квадратная')     
    for i in range(len(a)):
        summ = 0
        if len(a) == 1:
            summ += a[0][0]
            return summ
        else:
            for k in range(len(a)):
                b = copy.deepcopy(a)
                b.pop(0)
                for j in range(len(b)):
                    b[j].pop(k)
                summ += det(b)*((-1)**k)*int(a[0][k])
            return summ
        
        
def matrix_of_coefficients_is_square(a):
    """
    Функция показывает, является ли матрица квадртной.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    Returns
    -------
    bool
        True - если матрица квадратная; 
        False - иначе

    """
    if len(a)+1 == len(a[0]):
        return True        
    return False

def norma(matr):
    """
    Функция вычисляет бесконечную норму матрицы.

    Parameters
    ----------
    matr : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    Returns
    -------
    float
        Максимальная из сумм всех столбцов - норма матрицы

    """
    matr = np.array(matr)
    return matr.sum(axis=0).max()



def ab(matr):
    """
    Функция разбивает введенную пользователем матрицу на две.

    Parameters
    ----------
    matr : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    Returns
    -------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Квадратная матрица - левая часть введенной
    b : [float, float, ...]
        Правый столбец введенной матрицы

    """
    a = []
    b = []
    for i in range(len(matr)):
        a.append([])
        for j in range(len(matr[0])-1):
            a[i].append(matr[i][j])
    for i in range(len(matr)):
        b.append(matr[i][len(matr)])
    return a,b



def fix_diagonal(A):
    """
    Функция преобразовывает матрицу таким образом, чтобы на главной диагонали не было нулей.

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введённая пользователем

    Returns
    -------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Преобразованная матрица A без 0 на главной диагонали

    """
    for i in range(len(A)):
        if (A[i][i] == 0):
            for j in range(len(A)):
                if A[j][i] != 0:
                    for k in range(len(A) + 1):
                        A[i][k] -= A[j][k]
                    break
    return A



def iakobi(A):
    """
    Функция находит решение системы линейных уравнений методом Якоби.

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введённая пользователем

    Returns
    -------
    A: [[float, float, ...],
        [float, float, ...],
        ...]
        Матрица коэффициентов изначальной матрицы A
    x: [float, float, ...]
        Список решений системы
    [[float, float, ...],
     [float, float, ...],
     ...]
        Обратная матрица для матрицы A
        
    """    
    B = []
    for i in range(len(A)):
        B.append([])
        B[i].append(A[i][-1])

    B1 = []
    for i in range(len(A)):
        B1.append(A[i][-1])
    for i in range(len(A)):
        A[i].pop(-1)
    tempx = [0 for i in range(len(A))]
    e = float(input('Введите эпсилон: '))
    x=[0 for i in range(0,len(A))]
    flag = True
    count = 0 
    while flag == True or count>100:
        for i in range(len(A)): 
            tempx[i] = B1[i]
            count +=1
            for j in range(len(A[i])):
                if i!=j:
                    count +=1
                    tempx[i] -= A[i][j]*x[j]
            tempx[i]/=A[i][i]
            count +=1
        flag = False
        for k in range(len(A)):
            if abs(x[k]-tempx[k])>=e:
                flag = True
            x[k] = tempx[k]
        if count > 100:
            break
    else:
        return A, x, np.linalg.inv(A), False

    return A, x, np.linalg.inv(A), True



def jdmethod(a, b):
    """
    Функция находит решение системы линейных уравнений методом Жордана-Гаусса.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Квадратная матрица - левая часть изначальной
    b : [float, float, ...]
        Правый столбец изначальной матрицы

    Returns
    -------
    b : [float, float, ...]
        Список решений системы
    a_orig : [[float, float, ...],
             float, float, ...],
             ...]
        Изначальная квадратная матрица

    """
    a = np.array(a, float)
    a_orig = copy.deepcopy(a)
    b = np.array(b, float)
    n = len(b)
    for k in range(n):
        if np.fabs(a[k,k]) < 1.0e-12:
            for i in range(k+1,n):
                if np.fabs(a[i,k]) > np.fabs(a[k,k]):
                    for j in range(k,n):
                        a[k,j],a[i,j] = a[i,j],a[k,j]
                    b[k],b[i] = b[i],b[k]
                    break
        pivot = a[k,k]
        for j in range(k,n):
            a[k,j] /= pivot
        b[k] /= pivot
        for i in range(n):
            if i == k or a[i,k] == 0: continue
            factor = a[i,k]
            for j in range(k,n):
                a[i,j] -= factor * a[k,j]
            b[i] -= factor * b[k]
            
    return b, a_orig



def linang_inv_jdmethod(a):
    """
    Функция находит обратную матрицу для введённой с помощью метода Жордана-Гаусса.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Изначальная матрица

    Returns
    -------
    [[float, float, ...],
     [float, float, ...],
     ...]
        Обратная матрица для введённой

    """
    a = np.array(a, float)
    e = np.eye(len(a))
    a = np.concatenate((a, e), axis=1)
    n = len(a)
    m = len(a[0])
    for k in range(n):
        if np.fabs(a[k,k]) < 1.0e-12:
            for i in range(k+1,n):
                if np.fabs(a[i,k]) > np.fabs(a[k,k]):
                    for j in range(k,m):
                        a[k,j],a[i,j] = a[i,j],a[k,j]
                    break

        pivot = a[k,k]
        for j in range(k,m):
            a[k,j] /= pivot

        for i in range(n):
            if i == k or a[i,k] == 0: continue
            factor = a[i,k]
            for j in range(k,m):
                a[i,j] -= factor * a[k,j]

    return a[:, n:]



def fraction_jbmethod(a):
    """
    Функция находит решение системы линейных уравнений методом Жордана-Гаусса,
    используя тип данных fraction

    Parameters
    ----------
    a : [[float, float, ...],
        [float, float, ...],
        ...]
        Изначальная матрица

    Returns
    -------
    x :[float, float, ...]
        Список решений системы

    """
    # Создание массива numpy размера n и инициализация нулем для хранения вектора решения
    n = len(a)
    x = np.zeros(n)

    # Чтение коэффициентов расширенной матрицы
    for i in range(n):
        for j in range(n+1):
            a[i][j] = Fraction(a[i][j])

    # Применение Метода Жордана-Гаусса
    for k in range(n):
        if np.fabs(a[k,k]) < 1.0e-12:
            for i in range(k+1,n):
                if np.fabs(a[i,k]) > np.fabs(a[k,k]):
                    for j in range(k,n+1):
                        a[k,j],a[i,j] = a[i,j],a[k,j]
                    break

        for j in range(n):
            if k != j:
                ratio = a[j][k]/a[k][k]

                for l in range(n+1):
                    a[j][k] = a[j][l] - ratio * a[k][l]

    # Получение решения
    for i in range(n):
        x[i] = a[i][n]/a[i][i]

    return x


def isdegenerate(A):
    """
    Функция проверяет, есть ли решения у введённой системы уравнений.

    Parameters
    ----------
    A : [[float, float, ...],
        [float, float, ...],
        ...]
        Введённая матрица

    Returns
    -------
    bool
        False - матрица не является вырожденной, корни есть.
        True - матрица вырожденная, либо не является квадратной, корней быть не может.

    """
    # ВВОД МАТРИЦЫ И ПРОВЕРКА НА ВЫРОЖДЕННОСТЬ 
    matr = np.array(A)
        
    DELTA = 0.01
    
    flag = True  # решение существует
    if matrix_of_coefficients_is_square(matr):
        a = ab(matr)[0]
        numpy_det = np.linalg.det(a)
        if abs(numpy_det) <= DELTA:
            my_own_det = det(a)
            if abs(my_own_det) <= DELTA:
                flag = False
            else:
                flag = True
        else:
            flag = True
            
        if flag:
            return False
        else:
            return True #матрица вырожденная, единого решения не существует
    else:
        return True #Матрица коэффициентов не является квадратной


import csv
from scipy.fft import rfft, rfftfreq, irfft
from pywt import wavedec
import pywt
import numpy as np

def veivletfurie_read_csv(filename):
    
    """
    Функция для получения данных x, y координат из файла .csv

    Parameters
    ----------
    filename : str
        Абсолютный или локальный путь до файла .csv

    Returns
    -------
    my_list : [float, float, ...]
        Список координат x
    my_list2 : [float, float, ...]
        Список координат y

    """
    
    filename = input()
    with open(filename, 'r') as file:
        data = csv.reader(file, delimiter=',')
        data = list(data)
        my_list = []
        my_list2 = []
        for i in range(len(data)):
            my_list.append(eval(data[i][0]))
            my_list2.append(eval(data[i][1]))
    return my_list, my_list2



def custom_fft(x_list, y_list, do_ifft=False):
    """
    Функция БПФ (Быстрое Преобразование Фурье)

    Parameters
    ----------
    x_list : [float, float, ...]
        Координаты х исходного сигнала
    y_list : [float, float, ...]
        Координаты y исходного сигнала
    do_ifft : bool, optional
        True - функция вернёт так-же востановленный сигнал. 
        По умолчанию - False.

    Returns
    -------
    X : [float, float, ...]
        Список частот
    Y : [float, float, ...]
        Список коеффициентов частот
    Y1 : [float, float, ...], optional
        Список значений востановленного сигнала, 
        возвращается только при do_ifft == True
    """
    
    period = (len(x_list))/(x_list[-1] - x_list[0])
    X = rfftfreq(len(x_list), 1 / period)
    fft_y = rfft(y_list)
    Y = [abs(el) for el in fft_y]
    if do_ifft:
        ifft_y = irfft(fft_y)
        Y1 = [el.real for el in ifft_y]
        return X, Y, Y1
    return X, Y



def no_sin_coef(X, Y):
    """
    Функция вычисления коэффициента несинусоидальности

    Parameters
    ----------
    X : [float, float, ...]
        Список частот
    Y : [float, float, ...]
        Список коеффициентов частот

    Returns
    -------
    coef : float
        Коэффициент несинусоидальности сигнала

    """
    
    step = X[1] - X[0]
    step = X[1] - X[0]
    s0 = ((Y[Y.index(max(Y)) + 1] + Y[Y.index(max(Y)) - 1])/2 + Y[Y.index(max(Y))])*step
    si = 0
    for i in range(1, len(Y)):
        si += (Y[i-1] + Y[i])* step/2
    coef = (si-s0)/s0
    return coef



def dobeshi(y, level):
    """
    Функция вычисления декомпозиции на основе вейвлет преобразования Добеши

    Parameters
    ----------
    y : [float, float, ...]
        Список коеффициентов частот
    level : int
        Уровень декомпозиции

    Returns
    -------
    coef : [float, float, ...]
        Список значений докемпозиции

    """
    
    coeffs = wavedec(y, 'db5', level=level)[1]
    return coeffs



def haar(y, level):
    """
    Функция вычисления декомпозиции на основе вейвлет преобразования Хааре

    Parameters
    ----------
    y : [float, float, ...]
        Список коеффициентов частот
    level : int
        Уровень декомпозиции

    Returns
    -------
    coef : [float, float, ...]
        Список значений докемпозиции

    """
    
    coeffs = wavedec(y, 'haar', level=level)[1]
    return coeffs



def wavelets_4lvl_gm(signal, level=1):
    """
    

    Parameters
    ----------
    signal : [float, float, ...]
        Список коеффициентов частот
    level : int, optional
        Уровень декомпозиции. По умолчанию 1.
        Доступны лишь значения 1, 2, 3, 4

    Raises
    ------
    ValueError
        Возникает при передаче level не входящего в интервал [1; 4].

    Returns
    -------
    coef : [float, float, ...]
        Список значений докемпозиции

    """
    
    if level < 1 or level > 4:
        raise ValueError('Доступные уровни декомпозиции: 1, 2, 3, 4')
    wavelet = pywt.ContinuousWavelet('mexh')
    coef1, freqs1 = pywt.cwt(signal, np.arange(1, 30), wavelet)
    coef2, freqs2 = pywt.cwt(coef1[0], np.arange(1, 30), wavelet)
    coef3, freqs3 = pywt.cwt(coef2[0], np.arange(1, 30), wavelet)
    coef4, freqs4 = pywt.cwt(coef3[0], np.arange(1, 30), wavelet)
    coefs = [coef1, coef2, coef3, coef4]
    coef = coefs[level - 1]
    return coef


from nummethvera.approxinterp import *
from nummethvera.diffequat import *
from nummethvera.diffintegr import *
from nummethvera.matrixop import *
from nummethvera.otjigants import *
from nummethvera.slau import *
from nummethvera.veivletfurie import *

from . import approxinterp
from . import diffequat
from . import diffintegr
from . import matrixop
from . import otjigants
from . import slau
from . import veivletfurie

__version__ = '0.1.7'
__all__ = """
    approxinterp_read_csv
    polinom_chebyshev
    lagranje
    newton
    jdmethod_fast
    qubic_splines
    linear_approximation
    dispersion_linear
    quadratic_approximation
    dispersion_quadratic
    Gauss_approximation
    dispersion_Gauss
    slau_inp_matr
    csv_inp_matr
    slau_outp_matr
    det
    matrix_of_coefficients_is_square
    norma
    ab
    fix_diagonal
    iakobi
    jdmethod
    linang_inv_jdmethod
    fraction_jbmethod
    isdegenerate
    draw_plot
    draw_plots
    input_expression
    lambda_func
    parse_string
    eyler_cauchy
    runge_kutti
    eyler_cauchy_system
    runge_kutti_system
    solve_expressions
    input_func
    diff
    integ
    matrixop_inp_matr
    matrixop_outp_matr
    transp
    summfunc
    subtfunc
    multnumfunc
    multfunc
    process
    f
    collect_nodes
    draw_graph
    transfer_labels
    get_matrix
    ant_colony_algoritm
    otjig
    veivletfurie_read_csv
    custom_fft
    no_sin_coef
    dobeshi
    haar
    wavelets_4lvl_gm
    """.split()



import copy

def matrixop_inp_matr():
    """
    Функция возвращает матрицу, введённую пользователем с клавиатуры.

    Returns
    -------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    """
    while True:
        try:
            m = int(input('Сколько будет строк в матрице? '))
        except:
            print('Вы ввели не число')
        else:
            if m > 0:
                break
            else:
                print('Вы ввели не натуральное число') 

    while True:
        try:
            n = int(input('Сколько будет столбцов в матрице? '))
        except:
            print('Вы ввели не число')
        else:
            if n > 0:
                break
            else:
                print('Вы ввели не натуральное число')
                
    print("Введите элементы матрицы (заполнение идёт по строкам)")
    a = []
    for i in range(m):
        a.append([])
        for j in range(n):
            while True:
                try:
                    print(f'Введите элемент a[{i+1}][{j+1}]')
                    elem = eval(input())
                except:
                    print('Вы ввели не число')
                else:
                    break
            a[i].append(elem)
    return a



def matrixop_outp_matr(a):
    """
    Выводит на экран пользователя матрицу a.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    """
    for i in range(len(a)): 
        print(f'{a[i]}')      



def transp(A):
    """
    Возвращает транспонированную исходную матрицу

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, которую нужно транспонировать

    Returns
    -------
    list
        Транспонированная матрица

    """
    print('Транспонированная матрица:')
    return [[x[i] for x in A] for  i in range(len(A[0]))]



def summfunc(A,B):
    """
    Сложение двух матриц

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Первая матрица, которую нужно сложить
    B : [[float, float, ...],
         [float, float, ...],
         ...]
        Вторая матрица, которую нужно сложить

    Raises
    ------
    ValueError
        Возникает при разных размерностях матриц.

    Returns
    -------
    summ : [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является суммой матриц A и B.

    """
    n = len(A)
    m = len(A[0])
    n1 = len(B)
    m1 = len(B[0])
    if n != n1 or m != m1:
        raise ValueError('Разные размерности матриц, введите матрицу/ы заново')
    summ = []
    for i in range(n):
        summ.append([])
        for j in range(m):
            summ[i].append(A[i][j] + B[i][j])
    return summ



def subtfunc(A,B):
    """
    Вычитание одной матрицы из другой

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, из которой будем вычитать
    B : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, которую будем вычитать

    Raises
    ------
    ValueError
        Возникает при разных размерностях матриц.

    Returns
    -------
    subt : [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является разностью матриц A и B

    """
    n = len(A)
    m = len(A[0])
    n1 = len(B)
    m1 = len(B[0])
    if n != n1 or m != m1:
        raise ValueError('Разные размерности матриц, введите матрицу/ы заново')
    subt = []
    for i in range(n):
        subt.append([])
        for j in range(m):
            subt[i].append(A[i][j] - B[i][j])
    return subt



def multnumfunc(num, A):
    """
    Умножение матрицы на число или список.

    Parameters
    ----------
    num : float
        Число, на которое нужно умножить матрицу
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, которую умножаем на число

    Raises
    ------
    ValueError
        Возникает, если первым элементов является матрица, а вторым - число.

    Returns
    -------
    mult : [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является результатом умножения числа num на матрицу A.

    """
    if type(num) != type(1) and type(num) != type(1.1) or type(A) != type([]):
        raise ValueError('Первый элемент должен быть числом, а второй - матрицей')
    n = len(A)
    m = len(A[0])
    mult = []
    for i in range(n):
        mult.append([])
        for j in range(m):
            mult[i].append(A[i][j] * num)
    return mult



def multfunc(A,B):
    """
    Умножение двух матриц

    Parameters
    ----------
    A : [[float, float, ...],
         [float, float, ...],
         ...]
        Первая матрица, которую будем складывать
    B : [[float, float, ...],
         [float, float, ...],
         ...]
        Вторая матрица, которую будем складывать 

    Raises
    ------
    ValueError
        Возникает, когда кол-во строк матрицы A не совпадает с кол-вом столбцов матрицы B.

    Returns
    -------
    mult : [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является произведением матриц A и B

    """
    n = len(A)
    m = len(B[0])
    t = len(A[0])
    l = len(B)
    if t != l:
        raise ValueError('Кол-во столбцов первой матрицы и кол-во строк второй должны совпадать. Попробуйте ввести матрицы еще раз!')
    mult = []
    for i in range(n):
        mult.append([])
        for j in range(m):
            s = 0
            for k in range(t):
                s += A[i][k] * B[k][j]
            mult[i].append(s)
    return mult



def process(func):
    """
    Выводит преобразованную в команды строку

    Parameters
    ----------
    func : str
        Строка с командами, введённая пользователем

    Returns
    -------
    func : str
        Преобразованная в команды строка

    """
    dicti = {'@': 'multfunc', '*': 'multnumfunc', '+': 'summfunc', '-': 'subtfunc'}
    func = func.replace('(', ' ( ').replace(')', ' ) ')
    func = func.split()
    while '(' in func:
        i1 = func.index('(')
        count = 1
        for r in range(i1 + 1, len(func)):
            if func[r] == '(':
                count += 1
            elif func[r] == ')':
                count -= 1
            if count == 0:
                i2 = r
                break
        part1 = func[:i1]
        part2 = func[i2 + 1:]
        res = process(' '.join(func[i1 + 1:i2]))
        func = part1 + res + part2
    for _ in range(len(func)):
        i = '*'
        if i in func:
            index = func.index(i)
            l1 = func[index-1]
            l2 = func[index]
            l3 = func[index+1]
            out = dicti[l2] + '(' + l1 + ',' + l3 + ')'
            func[index] = out
            func.pop(index+1)
            func.pop(index-1)
                
    for _ in range(len(func)):
        i = '@'
        if i in func:
            index = func.index(i)
            l1 = func[index-1]
            l2 = func[index]
            l3 = func[index+1]
            out = dicti[l2] + '(' + l1 + ',' + l3 + ')'
            func[index] = out
            func.pop(index+1)
            func.pop(index-1)
            
    for _ in range(len(func)):
        for i in ["+", "-"]:
            if i in func:
                index = func.index(i)
                l1 = func[index-1]
                l2 = func[index]
                l3 = func[index+1]
                out = dicti[l2] + '(' + l1 + ',' + l3 + ')'
                func[index] = out
                func.pop(index+1)
                func.pop(index-1)
    return func



def f(func):
    """
    Функция выполняет все математические операции, введённые пользователем.

    Parameters
    ----------
    func : str
        Преобразованная в командры строка

    Returns
    -------
    [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является резульатом выполнения всех операций, введённых пользователем.

    """
    func = ''.join(process(func))
    print(func)
    return eval(func)



def det(a):
    """
    Функция вычисляет определитель матрицы.

    Parameters
    ----------
    a : [[float, float, ...],
         [float, float, ...],
         ...]
        Матрица, введенная пользователем

    Raises
    ------
    ValueError
        Возникает, если введённая матрица a не квадратная

    Returns
    -------
    summ : float
        Определитель матрицы

    """
    if len(a) != len(a[0]):
        raise ValueError('Матрица должна быть квадратная')     
    for i in range(len(a)):
        summ = 0
        if len(a) == 1:
            summ += a[0][0]
            return summ
        else:
            for k in range(len(a)):
                b = copy.deepcopy(a)
                b.pop(0)
                for j in range(len(b)):
                    b[j].pop(k)
                summ += det(b)*((-1)**k)*int(a[0][k])
            return summ


def calculation(func):
    """
    Функция, в которой выполняются нужные математические операции

    Parameters
    ----------
    func : str
        Функция преобразования матриц, введённая пользователем

    Returns
    -------
    [[float, float, ...],
            [float, float, ...],
            ...]
        Матрица, которая является результатом вычисления

    """
    while True:
        func = input('Введите преобразование: ')
        if func.count('(') != func.count(')'):
            print('Кол-во открывающих скобок должно быть равно кол-ву закрывающих скобок')
        else:
            break
    temp = func.replace('+', ' ').replace('-', ' ').replace('*', ' ').replace('@', ' ').replace(')', ' ').replace('(', ' ')
    temp = temp.split()
    temp = sorted(list(set(temp)))
    print(temp)
    for s in temp:
        if s.isalpha():
            print(f'Введите матрицу {s}')
            exec(f'{s} = inp_matr()')
            eval(f"outp_matr({s})")
    return f(func)







from setuptools import setup, find_packages

setup(name='nummethvera', 
      version='0.1.7', 
      url='https://github.com/dthseemsbttr/numericmethodsdths', 
      license='MIT', 
      author='Vera Korotkova',
      author_email='verakorotkova10@mail.ru', 
      description='Numeric methods library',
      packages=['nummethvera'],
      long_description=open('README.md').read(),
      setup_requires=['sympy>=1.6.2', 'mpmath>=1.1.0', 'scipy>=1.5.2',
                      'numpy>=1.19.2,<1.21', 'matplotlib>=3.5.0',
                      'PyWavelets>=1.1.1', 'numba>=0.51.2',
                      'networkx>=2.5'],
      zip_safe=False)


from math import *
import mpmath 

def input_func():
    """
    Возвращает введенную функцию и её параметры.

    Returns
    -------
    func : str
        Исходная функция
    a : float
        Левая граница интервала
    b : float
        Правая граница интервала
    h : float
        Шаг или точность

    """
    
    mpmath.mp.dps = 64
    func = input('Введите функцию\n')
    
    def f(x):
        z = eval(func)
        return z
    
    while True:
        try:
            f(1)
        except Exception as e:
            print(e)
            print('Ошибка в функции, введите функцию еще раз\n')
            func = input()
        else:
            break
    
    while True:
        try:
            a = float(input('Введите левую границу интервала\n'))
        except:
            print('Введите число')
        else:
            break
    
    while True:
        try:
            b = float(input('Введите правую границу интервала\n'))
        except:
            print('Введите число')
        else:
            break
    
    ans = input('Вы хотите задать шаг или точность? (1 - шаг; 2 - точность)\n')
    flag = True
    while flag: 
        if ans == '1':
            while True:
                try:
                    h = float(input('Введите шаг\n'))
                except:
                    print('Введите число, например 0.001')
                else:
                    break
            flag = not flag
        elif ans == '2':
            while True:
                try:
                    h = float(input('Введите точность\n'))**0.5
                except:
                    print('Введите число, например 0.001')
                else:
                    break
            flag = not flag
        else:
            print('Введите "1" или "2"')
            ans = input()
            
    if b < a:
        a, b = b, a
        
    return func, a, b, h



def diff(a, b, h, func):
    """
    Дифференцирование по формуле

    Parameters
    ----------
    a : float
        Левая граница интервала
    b : float
        Правая граница интервала
    h : float
        Шаг или точность
    func : str
        Исходная функция

    Returns
    -------
    xspis : [float, float, ...]
        Список значений координат x
    yspis : [float, float, ...]
        Список значений координат y
    ydiffspis : [float, float, ...]
        Список значений производной функции 

    """
    ydiffspis = []
    yspis = []
    xspis = []
    banspis = []
    
    def f(x):
        z = eval(func)
        return z
    
    c = int(1/h)
    for i in range(int(a*c),int(b*c)+1,round(h*c)):
        try:
            diff = (f((i/c)+h)-f((i/c)-h))/(2*h)
            yspis.append(f(i/c))
            xspis.append(i/c)
            ydiffspis.append(diff)
        except:
            print(f'Точка разрыва в координате х = {i/c}')
            banspis.append(i/c)
            continue
    return xspis, yspis, ydiffspis



def integ(a, b, h, func):
    """
    Интегрирование по формуле

    Parameters
    ----------
    a : float
        Левая граница интервала
    b : float
        Правая граница интервала
    h : float
        Шаг или точность
    func : str
        Исходная функция

    Returns
    -------
    S1 : float
        Значение определенного интеграла
    slist : [float, float, ...]
        История изменений значения определенного интеграла.

    """
    
    ylist1 = []
    slist = []
    xlist1 = []
    banlist = []
    
    def f(x):
        z = eval(func)
        return z

    S1 = 0
    c = int(1/h)
    for i in range(int(a*c),int(b*c),round(h*c)):
        try:
            S1 = S1 + ((f(i/c)+f(i/c+h))/2)*h
            xlist1.append(i/c)
            ylist1.append(f(i/c))
            slist.append(S1)
        except:
            print('Деление на 0')
            banlist.append(i/c)
            
    return S1, slist


import csv
import numpy as np
from numba import prange
import networkx as nx
import matplotlib.pyplot as plt
from functools import reduce


def collect_nodes(points, highlighted_nodes=[], color='#79FF06', highlighted_color='blue', width=200, highlighted_width=400):
    """
    Собирает необходимые нам вершины в нужный формат

    Parameters
    ----------
    points : [str, str, ...]
        Вершины графа.
    highlighted_nodes : [str, str, ...], optional
        Выделенные вершины графа. 
        По умолчанию [].
    color : str, optional
        Цвет обычных вершин. 
        По умолчанию '#79FF06'.
    highlighted_color : str, optional
        Цвет выделенных вершин. 
        По умолчанию 'blue'.
    width : int, optional
        Ширина обычных вершин. 
        По умолчанию 200.
    highlighted_width : int, optional
        Ширина выделенных вершин. 
        По умолчанию 400.

    Returns
    -------
    result : [(str, {'color': str, 'width': int}),
              (str, {'color': str, 'width': int}),
              ...]
        Список вершин с их параметрами.

    """
    result = []
    for p in points:
        if p in highlighted_nodes:
            result.append((p, {"color": highlighted_color, "width": highlighted_width}))
        else:
            result.append((p, {"color": color, "width": width}))
    return result
    

def draw_graph(points, labels, dict_labels, highlighted_nodes=[], with_labels=False, with_label_labels=False):
    """
    Функция, рисующая заданный граф

    Parameters
    ----------
    points : [str, str, ...]
        Вершины графа.
    labels : [(str, str, int, str),
              (str, str, int, str),
              ...]
        Список кортежей из рёбер типа (<Вершина1>, <Вершина2>, <Вес ребра>, <Цвет ребра>).
    dict_labels : {(str, str): int,
                   (str, str): int,
                   ...}
        Словарь рёбер с их весами.
    highlighted_nodes : [str, str, ...], optional
        Выделенные вершины графа. 
        По умолчанию [].
    with_labels : bool, optional
        Для вывода весов рёбер. 
        По умолчанию False.
    with_label_labels : bool, optional
        Для вывода названий вершин. 
        По умолчанию False.

    """
    
    graph = nx.Graph()
    args_points = collect_nodes(points, highlighted_nodes)
    graph.add_nodes_from(args_points)
    for label in labels:
        if label[3] == 'red':
            graph.add_edge(label[0],
                       label[1],
                       color=label[3],
                       weight=label[2],
                       width=7)
        else:
            graph.add_edge(label[0],
                       label[1],
                       color=label[3],
                       weight=label[2],
                       width=3)
    pos = nx.kamada_kawai_layout(graph)
    edges = graph.edges()
    nodes = graph.nodes
    node_colors = [graph.nodes[n]['color'] for n in nodes]
    node_widths = [graph.nodes[n]['width'] for n in nodes]
    colors = [graph[e[0]][e[1]]['color'] for e in edges]
    widths = [graph[e[0]][e[1]]['width'] for e in edges]
    plt.figure(1,figsize=(10,10))
    nx.draw(graph, pos, edgelist=edges, edge_color=colors, node_color=node_colors, with_labels=with_labels, width=widths, node_size=node_widths)
    if with_label_labels:
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=dict_labels)
    plt.show()

def transfer_labels(matrix, shortest_path, color='black', highlighted_color='red'):
    """
    Переводит матрицу весовых коэфициентов в список рёбер и вершин с их аттрибутами

    Parameters
    ----------
    matrix : [[float, float, ...],
              [float, float, ...],
              ...]
        Матрица весовых коэфициентов.
    shortest_path : [(str, str), (str, str), ...]
        Список из всех ребёр, формирующих кратчайший путь.
    color : str, optional
        Цвет обычных рёбер. 
        По умолчанию 'black'.
    highlighted_color : str, optional
        Цвет выделенных рёбер (из списка кратчайшего пути). 
        По умолчанию 'red'.

    Returns
    -------
    points : [str, str, ...]
        Вершины графа.
    labels : [(str, str, int, str),
              (str, str, int, str),
              ...]
        Список кортежей из рёбер типа (<Вершина1>, <Вершина2>, <Вес ребра>, <Цвет ребра>).
    dict_labels : {(str, str): int,
                   (str, str): int,
                   ...}
        Словарь рёбер с их весами.

    """
    labels = []
    dict_labels = {}
    points = [str(x) for x in range(len(matrix))]
    for i in range(len(matrix) - 1):
        for j in range(i + 1, len(matrix)):
            if tuple([i, j]) in shortest_path or tuple([j, i]) in shortest_path:
                labels.append((str(i), str(j), matrix[i, j], highlighted_color))
                labels.append((str(j), str(i), matrix[j, i], highlighted_color))
            else:
                labels.append((str(i), str(j), matrix[i, j], color))
                labels.append((str(j), str(i), matrix[j, i], color))
            dict_labels[(str(i), str(j))] = matrix[i, j]
            dict_labels[(str(j), str(i))] = matrix[j, i]
    return points, labels, dict_labels

def get_matrix():
    """
    Функция считывания или генерирования матрицы весовых коэффициентов

    Returns
    -------
    matrix : [[float, float, ...],
              [float, float, ...],
              ...]
        Матрица весовых коэфициентов.

    """
    while True:
        ans = input('Как вы хотите ввести матрицу весовых коэффициентов?\n 1. csv файл\n 2. Рандомная генерация\n 3. Ручной ввод (не рекомендуется при большом кол-ве вершин в графе)\n ->')
        if ans in ['1', '2', '3']:
            break
        print('Вы ввели не верный вариант, выберите пожалуйста из вариантов 1/2/3')

    if ans == '1':
        while True:
            filename = input('Введите название файла .csv\n ->')
            try:
                with open(filename, 'r') as file:
                    pass
            except FileNotFoundError:
                print('Такой файл не найден, попробуйте еще раз')
                continue
            if filename.endswith('.csv'):
                break
            print('Вы ввели файл с отличным от .csv расширением, попробуйте еще раз')


        with open(filename, 'r') as file:
            matrix = []
            data = csv.reader(file)
            for row in data:
                matrix.append([float(element) for element in row])
            matrix = np.array(matrix)
    elif ans == '2':
        n = int(input('Введите кол-во городов (вершин)\n ->'))
        matrix = np.zeros((n, n), float)
        np.fill_diagonal(matrix, float('inf'))
        for i in range(n - 1):
            for j in range(i + 1, n):
                value = round(np.random.uniform() * (n - 1) + 1, 3)
                matrix[i][j] = value
                matrix[j][i] = value
    else:
        n = int(input('Введите кол-во городов (вершин)\n ->'))
        matrix = np.zeros((n, n), float)
        np.fill_diagonal(matrix, float('inf'))
        for i in range(n - 1):
            for j in range(i + 1, n):
                value = float(input(f'Введите длину дороги (вес ребра) между городами {i} и {j}: '))
                matrix[i][j] = value
                matrix[j][i] = value
    return matrix, len(matrix)


def ant_colony_algoritm(distances, decay=0.95, alpha=1, beta=1, epsilon=0.001):
    """
    Оптимизация методом имитации муравьиной колонии

    Parameters
    ----------
    distances : [[float, float, ...],
              [float, float, ...],
              ...]
        Матрица весовых коэфициентов.
    decay : float, optional
        Коэф выветривания феромонов. Строго меньше 1.
        По умолчанию 0.95.
    alpha : int, optional
        Коэф. значимости феромонов. 
        По умолчанию 1.
    beta : int, optional
        Коэф значимости растояния. 
        По умолчанию 1.
    epsilon : float, optional
        Точность вычисления. Рекомендуется брать малые значения
        По умолчанию 0.001.

    Returns
    -------
    all_time_shortest_path : ((str, str),
                              (str, str),
                              ...,
                              int)
        Кортеж из рёбер, составлящий кратчайший путь и вес этого пути.
    n_iterations : int
        Кол-во иттераций, потребовавшихся для завершения работы.
    paths_list : [float, float, ...]
        Список всех изменений веса кратчайшего пути за время работы программы.

    """
    
    def gen_all_paths():
        all_paths = []
        for i in prange(n_ants):
            path = gen_path(i)
            all_paths.append((path, gen_path_dist(path)))
        return all_paths
    
    def gen_path_dist(path):
        total_dist = 0
        for ele in path:
            total_dist += distances[ele]
        return total_dist
    
    def gen_path(start):
        path = []
        visited = set()
        visited.add(start)
        prev = start
        for i in prange(len(distances) - 1):
            move = pick_move(pheromone[prev], distances[prev], visited)
            path.append((prev, move))
            prev = move
            visited.add(move)
        path.append((prev, start))   
        return path
    
    def pick_move(pheromone, dist, visited):
        pheromone = np.copy(pheromone)
        pheromone[list(visited)] = 0

        row = pheromone ** alpha * (( 1.0 / dist) ** beta)

        norm_row = row / row.sum()
        move = np.random.choice(all_inds, 1, p=norm_row)[0]
        return move
    
    def spread_pheronome(all_paths):
        sorted_paths = sorted(all_paths, key=lambda x: x[1])
        for path, dist in sorted_paths:
            for move in path:
                pheromone[move] += 1.0 / distances[move]

    n_ants=len(distances)
    n_iterations=len(distances) ** 2 // 10
    pheromone = np.ones(distances.shape) / len(distances)
    all_inds = range(len(distances))
    shortest_path = None
    all_time_shortest_path = ("", np.inf)
    same_count = 0
    paths_list = []
    for i in prange(n_iterations):
        all_paths = gen_all_paths()
        spread_pheronome(all_paths)
        shortest_path = min(all_paths, key=lambda x: x[1])
        if shortest_path[1] == all_time_shortest_path[1]:
            same_count += 1
        else:
            same_count = 0
        if shortest_path[1] < all_time_shortest_path[1]:
            all_time_shortest_path = shortest_path 
        paths_list.append(all_time_shortest_path[1])
        if same_count == 5:
            break
        pheromone = pheromone * decay
    return all_time_shortest_path, n_iterations, paths_list


def otjig(matrix, t_min=0.5, t_max=1000):
    """
    Алгоритм оптимизации имитацией отжига

    Parameters
    ----------
    matrix : [[float, float, ...],
              [float, float, ...],
              ...]
        Матрица весовых коэфициентов.
    t_min : float, optional
        Минимальная температура. 
        По умолчанию 0.5.
    t_max : float, optional
        Максимальная температура. 
        По умолчанию 1000.

    Returns
    -------
    result : ((str, str),
                              (str, str),
                              ...,
                              int)
        Кортеж из рёбер, составлящий кратчайший путь и вес этого пути.
    count : int
        Кол-во иттераций, потребовавшихся для завершения работы.
    paths_list : [float, float, ...]
        Список всех изменений веса кратчайшего пути за время работы программы.

    """
    n = len(matrix)
    T_0 = np.array(np.linspace(0, n-1, n), int)
    np.random.shuffle(T_0)
    v = sum([matrix[T_0[i-1]][T_0[i]] for i in range(len(T_0))])
    temp = t_max
    iteration = 0
    count = 0
    paths_list = []
    while temp > t_min:
        count += 1
        i, j = sorted(np.random.choice(np.array(np.linspace(1, n-1, n-1), int),  size=2, replace=False))
        T = T_0.copy()
        T[i:j+1] = T[j:i-1:-1]
        v_new = sum([matrix[T[k-1]][T[k]] for k in range(len(T))])
        paths_list.append(v)
        if v_new < v:
            iteration += 1
            v = v_new
            T_0 = T
            temp = t_max / iteration
        elif v < v_new:
            if np.random.uniform() <= np.exp(-(v_new - v)/temp):
                iteration += 1
                v = v_new
                T_0 = T
                temp = t_max / iteration
    result = []
    for i in range(len(T_0)):
        result.append((T_0[i - 1], T_0[i]))     
                
    return (result, v), count, paths_list